#!/usr/bin/env python

from cStringIO import StringIO
import unittest
from unittest import TestCase

import ClientForm

# XXX
# ParseTests should use FormParser directly, and check output directly, rather
#  than via HTMLForm
# Form.enctype

class LWPFormTests(TestCase):
    """The original tests from libwww-perl 5.64."""
    def testEmptyParse(self):
        forms = ClientForm.ParseFile(StringIO(""), "http://localhost")
        assert len(forms) == 0

    def _forms(self):
        file = StringIO("""<form action="abc">

        <input name="firstname" value="Gisle">

        </form>

        """)
        return ClientForm.ParseFile(file, "http://localhost/")

    def testParse(self):
        forms = self._forms()
        assert len(forms) == 1
        assert forms[0]["firstname"] == "Gisle"

    def testFillForm(self):
        forms = self._forms()
        form = forms[0]
        form["firstname"] = "Gisle Aas"
        req = form.click()
        def request_method(req):
            if req.has_data():
                return "POST"
            else:
                return "GET"
        assert request_method(req) == "GET"
        assert req.get_full_url() == "http://localhost/abc?firstname=Gisle+Aas"


class ParseTests(TestCase):
    # Remember to move test for form.name somewhere else if rehash these tests
    # so they don't use HTMLForm.
    def testTextarea(self):
        file = StringIO(
"""<form action="abc">

<input name="firstname" value="Gisle">
<textarea>blah, blah,
Rhubarb.

</textarea>

<textarea></textarea>

<textarea name="ta">Hello testers &amp; users!</textarea>

</form>

""")
        forms = ClientForm.ParseFile(file, "http://localhost/")
        assert len(forms) == 1
        form = forms[0]
        assert form.name is None
        control = form.find_control(type="textarea")
        assert control.name is None
        assert control.value == "blah, blah,\nRhubarb.\n\n"

        empty_control = form.find_control(type="textarea", nr=1)
        assert str(empty_control) == "<TextareaControl(<None>=)>"
        assert empty_control.value == ""

        entity_ctl = form.find_control(type="textarea", nr=2)
        self.assert_(entity_ctl.name == "ta")
        self.assert_(entity_ctl.value == "Hello testers & users!")

    def testSelect(self):
        file = StringIO(
"""<form action="abc">

<select name="foo">
 <option>Hello testers &amp; users!</option>
 <option></option><option></option>
</select>

</form>

""")
        forms = ClientForm.ParseFile(file, "http://localhost/")
        self.assert_(len(forms) == 1)
        form = forms[0]

        entity_ctl = form.find_control(type="select")
        self.assert_(entity_ctl.name == "foo")
        self.assert_(entity_ctl.value == "Hello testers & users!")
        self.assert_(entity_ctl.get_value_as_label() ==
                     "Hello testers & users!")

    def testButton(self):
        file = StringIO(
"""<form action="abc" name="myform">

<input type="text" value="cow" name="moo">

<button name="b">blah, blah,
Rhubarb.</button>

<button type="reset" name="b2"></button>
<button type="button" name="b3"></button>

</form>

""")
        forms = ClientForm.ParseFile(file, "http://localhost/")
        form = forms[0]
        assert form.name == "myform"
        control = form.find_control(name="b")
        assert control.type == "submitbutton"
        assert control.value is None
        items = form.click(items=1)
        assert items == [("moo", "cow")]

    def testIsindex(self):
        file = StringIO(
"""<form action="abc">

<isindex prompt=">>>">

</form>

""")
        forms = ClientForm.ParseFile(file, "http://localhost/")
        form = forms[0]
        control = form.find_control(type="isindex")
        assert control.type == "isindex"
        assert control.name is None
        assert control.value == ""
        control.value = "some stuff"
        assert control.isindex_url() == "?some+stuff"
        items = form.click(items=1)
        assert items == []

# known bug, will probably fix in 0.1.0
##     def testEmptySelect(self):
##         file = StringIO(
## """<form action="abc">
## <select></select>

## <select multiple></select>

## </form>
## """)
##         forms = ClientForm.ParseFile(file, "http://localhost/")
##         form = forms[0]
##         control0 = form.find_control(type="select", nr=0)
##         control1 = form.find_control(type="select", nr=1)
##         print "str(control0)", str(control0)
##         assert str(control0) == str(control1) == "SelectControl()"

    def testEmptyListControls(self):
        file = StringIO("""<form action="./weird.html">

<input type="checkbox" name="foo"></input>

<input type="radio" name="bar"></input>

<!--
<select name="baz">
  <option></option>
</select>

<select name="baz" multiple>
  <option></option>
</select>
-->

<input type="submit" name="submit">
</form>
""")
        forms = ClientForm.ParseFile(file, "http://localhost/")
        form = forms[0]
        assert form.possible_values("foo") == ["on"]
        assert form.possible_values("bar") == ["on"]
        #assert form.possible_values("baz") == []
        assert form["foo"] == []
        assert form["bar"] == None
        #assert form["baz"] == []
        form["foo"] = "on"
        form["bar"] = "on"
        items = form.items()
        assert items == [("foo", "on"), ("bar", "on")]

    def testBadSingleSelect(self):
        # HTML 4.01 section 17.6.1: single selection SELECT controls shouldn't
        # have > 1 item selected, but if they do, not more than one should end
        # up selected.
        file = StringIO("""<form action="./bad.html">

<select name="spam">
  <option selected>1</option>
  <option selected>2</option>
</select>

<input type="submit" name="submit">
</form>
""")
        forms = ClientForm.ParseFile(file, "http://localhost/")
        form = forms[0]
        assert form.possible_values("spam") == ["1", "2"]
        nr_selected = len(form.find_control("spam").items())
        assert nr_selected == 1

    def testSelectDefault(self):
        file = StringIO(
"""<form action="abc" name="myform">

<select name="a" multiple>
 <option>1</option>
 <option>2</option>
 <option>3</option>
</select>

<select name="b">
 <option>1</option>
 <option>2</option>
 <option>3</option>
</select>

</form>

""")
        forms = ClientForm.ParseFile(file, "http://localhost/")
        form = forms[0]
        control = form.find_control("a")
        assert control.value == []
        single_control = form.find_control("b")
        assert single_control.value == "1"

        file.seek(0)
        forms = ClientForm.ParseFile(file, "http://localhost/",
                                     select_default=1)
        form = forms[0]
        # select_default only affects *multiple* selection select controls
        control = form.find_control(type="select")
        assert control.value == ["1"]
        single_control = form.find_control(type="select", nr=1)
        assert single_control.value == "1"
##     def testSelectDefault(self):
##         file = StringIO(
## """<form action="abc" name="myform">

## <select multiple>
##  <option>1</option>
##  <option>2</option>
##  <option>3</option>
## </select>

## </form>

## """)
##         forms = ClientForm.ParseFile(file, "http://localhost/")
##         form = forms[0]
##         control = form.find_control(type="select")
##         assert control.value == []

##         file.seek(0)
##         forms = ClientForm.ParseFile(file, "http://localhost/",
##                                      select_default=1)
##         form = forms[0]
##         control = form.find_control(type="select")
##         assert control.value == ["1"]


class ControlTests(TestCase):
    def testTextControl(self):
        attrs = {"type": "this is ignored",
                 "name": "ath_Uname",
                 "value": "",
                 "maxlength": "20"}
        c = ClientForm.TextControl("texT", attrs)
        c.fixup()
        assert c.type == "text"
        assert c.name == "ath_Uname"
        assert c.value == ""
        assert str(c) == "<TextControl(ath_Uname=)>"
        assert c.possible_values() == []
        assert c.items() == [("ath_Uname", "")]
        def bad_assign(c=c): c.type = "sometype"
        self.assertRaises(AttributeError, bad_assign)
        assert c.type == "text"
        def bad_assign(c=c): c.name = "somename"
        self.assertRaises(AttributeError, bad_assign)
        assert c.name == "ath_Uname"
        c.value = "2"
        assert c.value == "2"
        assert str(c) == "<TextControl(ath_Uname=2)>"
        assert not c.readonly
        c.readonly = 1
        def bad_assign(c=c): c.value = "foo"
        self.assertRaises(AttributeError, bad_assign)
        assert c.value == "2"
        c.disabled = 1
        assert str(c) == "<TextControl(ath_Uname=2) (disabled, readonly)>"
        c.readonly = 0
        assert str(c) == "<TextControl(ath_Uname=2) (disabled)>"
        self.assertRaises(AttributeError, bad_assign)
        assert c.value == "2"
        assert c.items() == []
        c.disabled = 0
        assert str(c) == "<TextControl(ath_Uname=2)>"

        assert c.attrs.has_key("maxlength")
        for key in "name", "type", "value":
            assert not c.attrs.has_key(key)

        # initialisation of readonly and disabled attributes
        attrs["readonly"] = 1
        c = ClientForm.TextControl("text", attrs)
        def bad_assign(c=c): c.value = "foo"
        self.assertRaises(AttributeError, bad_assign)
        del attrs["readonly"]
        attrs["disabled"] = 1
        c = ClientForm.TextControl("text", attrs)
        def bad_assign(c=c): c.value = "foo"
        self.assertRaises(AttributeError, bad_assign)
        del attrs["disabled"]
        c = ClientForm.TextControl("hidden", attrs)
        assert c.readonly
        def bad_assign(c=c): c.value = "foo"
        self.assertRaises(AttributeError, bad_assign)

    def testIsindexControl(self):
        attrs = {"type": "this is ignored",
                 "prompt": ">>>"}
        c = ClientForm.IsindexControl("isIndex", attrs)
        c.fixup()
        assert c.type == "isindex"
        assert c.name is None
        assert c.value == ""
        assert str(c) == "<IsindexControl()>"
        assert c.possible_values() == []
        assert c.items() == []
        def bad_assign(c=c): c.type = "sometype"
        self.assertRaises(AttributeError, bad_assign)
        assert c.type == "isindex"
        def bad_assign(c=c): c.name = "somename"
        self.assertRaises(AttributeError, bad_assign)
        assert c.name is None
        c.value = "2"
        assert c.value == "2"
        assert str(c) == "<IsindexControl(2)>"
        c.disabled = 1
        assert str(c) == "<IsindexControl(2) (disabled)>"
        def bad_assign(c=c): c.value = "foo"
        self.assertRaises(AttributeError, bad_assign)
        assert c.value == "2"
        assert c.items() == []
        c.disabled = 0
        assert str(c) == "<IsindexControl(2)>"

        assert c.attrs.has_key("prompt")
        assert c.attrs["prompt"] == ">>>"
        for key in "name", "type", "value":
            assert not c.attrs.has_key(key)

        c.value = "foo 1 bar 2"
        assert c.isindex_url() == "?foo+1+bar+2"

    def testIgnoreControl(self):
        attrs = {"type": "this is ignored"}
        c = ClientForm.IgnoreControl("reset", attrs)
        assert c.type == "reset"
        assert c.value == None
        assert str(c) == "<IgnoreControl(<None>=<None>)>"
        assert c.possible_values() == []

        def bad_assign(c=c): c.value = "foo"
        self.assertRaises(AttributeError, bad_assign)
        assert c.value == None

    def testSubmitControl(self):
        attrs = {"type": "this is ignored",
                 "name": "name_value",
                 "value": "value_value",
                 "img": "foo.gif"}
        c = ClientForm.SubmitControl("submit", attrs)
        assert c.type == "submit"
        assert c.name == "name_value"
        assert c.value == "value_value"
        assert str(c) == "<SubmitControl(name_value=value_value)>"
        assert c.possible_values() == []
        c.disabled = 1
        def bad_assign(c=c): c.value = "foo"
        self.assertRaises(AttributeError, bad_assign)
        assert str(c) == "<SubmitControl(name_value=value_value) (disabled)>"
        c.disabled = 0

        # click on button
        form = ClientForm.HTMLForm("http://foo.bar.com/")
        c.add_to_form(form)
        assert c.items() == []
        items = c.click(form, items=1)
        request = c.click(form)
        assert c.items() == []
        assert items == [("name_value", "value_value")]
        assert request.get_full_url() == \
               "http://foo.bar.com/?name_value=value_value"
        c.disabled = 1
        items = c.click(form, items=1)
        request = c.click(form)
        assert items == []
        assert request.get_full_url() == \
               "http://foo.bar.com/?"

    def testImageControl(self):
        attrs = {"type": "this is ignored",
                 "name": "name_value",
                 "value": "value_value",
                 "img": "foo.gif"}
        c = ClientForm.ImageControl("image", attrs)
        assert c.type == "image"
        assert c.name == "name_value"
        assert c.value == "value_value"
        assert str(c) == "<ImageControl(name_value=value_value)>"
        assert c.possible_values() == []

        # click, at coordinate (0, 55), on image
        form = ClientForm.HTMLForm("http://foo.bar.com/")
        c.add_to_form(form)
        assert c.items() == []
        request = c.click(form, (0, 55))
        assert c.items() == []
        assert request.get_full_url() == \
               "http://foo.bar.com/?name_value.x=0&name_value.y=55"
        c.disabled = 1
        assert c.click(form, items=1) == []
        c.disabled = c.readonly = 0
        self.assert_(c.click(form, (1,1), items=1) ==
                     [("name_value.x", "1"), ("name_value.y", "1")])

    def testCheckboxControl(self):
        attrs = {"type": "this is ignored",
                 "name": "name_value",
                 "value": "value_value",
                 "alt": "some string"}
        c = ClientForm.CheckboxControl("checkbox", attrs)
        c.fixup()
        assert c.type == "checkbox"
        assert c.name == "name_value"
        assert c.value == []
        assert c.possible_values() == ["value_value"]
        def bad_assign(c=c): c.type = "sometype"
        self.assertRaises(AttributeError, bad_assign)
        assert c.type == "checkbox"
        def bad_assign(c=c): c.name = "somename"
        self.assertRaises(AttributeError, bad_assign)
        assert c.name == "name_value"

        # construct larger list from length-1 lists
        c = ClientForm.CheckboxControl("checkbox", attrs)
        attrs2 = attrs.copy()
        attrs2["value"] = "value_value2"
        c2 = ClientForm.CheckboxControl("checkbox", attrs2)
        c.merge_control(c2)
        c.fixup()
        assert str(c) == ("<CheckboxControl("
                           "name_value=[value_value, value_value2])>")
        assert c.possible_values() == ["value_value", "value_value2"]

        attrs = c.find_item("value_value")
        assert attrs.has_key("alt")
        for key in "name", "value", "type":
            assert not attrs.has_key(key)
        assert c.find_item("oops") is None

        c.value = ["value_value", "value_value2"]
        assert c.value == ["value_value", "value_value2"]
        c.value = ["value_value"]
        assert c.value == ["value_value"]
        c.value = "value_value2"
        assert c.value == ["value_value2"]
        def bad_assign(c=c): c.value = "oops"
        self.assertRaises(AttributeError, bad_assign)
        assert c.value == ["value_value2"]
        c.toggle("value_value")
        assert c.value == ["value_value", "value_value2"]
        c.toggle("value_value2")
        assert c.value == ["value_value"]
        self.assertRaises(ValueError, c.toggle, "oops")
        assert c.value == ["value_value"]

        assert str(c) == ("<CheckboxControl("
                           "name_value=[*value_value, value_value2])>")
        c.disabled = 1
        def bad_assign(c=c): c.value = ["value_value"]
        assert str(c) == ("<CheckboxControl("
                           "name_value=[*value_value, value_value2]) "
                           "(disabled)>")
        assert c.value == ["value_value"]
        assert c.items() == []
        c.disabled = 0
        assert c.items() == [("name_value", "value_value")]
        c.value = None
        assert c.value == []

    def testSelectControlMultiple(self):
        attrs = {"type": "this is ignored",
                 "name": "name_value",
                 "value": "value_value",
                 "multiple": "",
                 "alt": "some string",
                 "label": "contents_value",
                 "contents": "contents_value"}
        # with Netscape / IE default selection...
        c = ClientForm.SelectControl("select", attrs)
        c.fixup()
        assert c.type == "select"
        assert c.name == "name_value"
        assert c.value == []
        assert c.possible_values() == ["value_value"]
        # ... and with RFC 1866 default selection
        c = ClientForm.SelectControl("select", attrs, select_default=1)
        c.fixup()
        assert c.value == ["value_value"]

        # construct larger list from length-1 lists
        c = ClientForm.SelectControl("select", attrs)
        attrs2 = attrs.copy()
        attrs2["value"] = "value_value2"
        c2 = ClientForm.SelectControl("select", attrs2)
        c.merge_control(c2)
        c.fixup()
        assert str(c) == ("<SelectControl("
                           "name_value=[value_value, value_value2])>")
        assert c.possible_values() == ["value_value", "value_value2"]

        # find_item
        attrs3 = c.find_item("value_value")
        assert attrs3.has_key("alt")
        # HTML attributes dictionary should have been copied by ListControl
        # constructor.  HTML attributes accessible through ListControl
        # interface should not be present in dictionary.
        attrs["new_attr"] = "new"
        attrs2["new_attr2"] = "new2"
        for key in ("name", "value", "type", "multiple",
                    "new_attr", "new_attr2"):
            assert not attrs3.has_key(key)
        assert c.find_item("oops") is None

        c.value = ["value_value", "value_value2"]
        assert c.value == ["value_value", "value_value2"]
        c.value = ["value_value"]
        assert c.value == ["value_value"]
        c.value = "value_value2"
        assert c.value == ["value_value2"]
        def bad_assign(c=c): c.value = "oops"
        self.assertRaises(AttributeError, bad_assign)
        assert c.value == ["value_value2"]
        c.toggle("value_value")
        assert c.value == ["value_value", "value_value2"]
        c.toggle("value_value2")
        assert c.value == ["value_value"]
        self.assertRaises(ValueError, c.toggle, "oops")
        assert c.value == ["value_value"]
        # test ordering of items
        c.value = ["value_value2", "value_value"]
        assert c.value == ["value_value", "value_value2"]
        c.value = None
        assert c.value == []

    def testSelectControlMultiple_label(self):
        import ClientForm
##         <SELECT name=year>
##          <OPTION value=0 label="2002">current year</OPTION>
##          <OPTION value=1>2001</OPTION>
##          <OPTION>2000</OPTION>
##         </SELECT>
        attrs = {"type": "ignored",
                 "name": "year",
                 "value": "0",
                 "multiple": "",
                 "label": "2002",
                 "contents": "current year"}
        attrs2 = {"type": "ignored",
                  "name": "year",
                  "value": "1",
                  "multiple": "",
                  "label": "2001",  # label defaults to contents
                  "contents": "2001"}
        attrs3 = {"type": "ignored",
                  "name": "year",
                  "value": "2000",  # value defaults to contents
                  "multiple": "",
                  "label": "2000",  # label defaults to contents
                  "contents": "2000"}
        c = ClientForm.SelectControl("select", attrs)
        c2 = ClientForm.SelectControl("select", attrs2)
        c3 = ClientForm.SelectControl("select", attrs3)
        c.merge_control(c2)
        c.merge_control(c3)
        c.fixup()

        assert c.possible_values() == ["0", "1", "2000"]
        assert c.possible_labels() == ["2002", "2001", "2000"]

        assert c.value == []
        c.toggle_by_label("2002")
        assert c.value == ["0"]
        c.toggle("0")
        assert c.value == []
        c.toggle("0")
        assert c.value == ["0"]
        assert c.get_value_as_label() == ["2002"]
        c.toggle_by_label("2002")
        self.assertRaises(ValueError, c.toggle_by_label, "blah")
        assert c.value == []
        c.toggle("2000")
        assert c.value == ["2000"]
        assert c.get_value_as_label() == ["2000"]

        def bad_value(c=c): c.value = "2002"
        assert c.value == ["2000"]
        c.value = "0"
        assert c.value == ["0"]
        c.value = []
        c.set_value_by_label("2002")
        assert c.value == ["0"]
        assert c.get_value_as_label() == ["2002"]
        c.set_value_by_label(["2000"])
        assert c.value == ["2000"]
        assert c.get_value_as_label() == ["2000"]
        c.set_value_by_label(["2000", "2002"])
        assert c.value == ["0", "2000"]
        assert c.get_value_as_label() == ["2002", "2000"]

    def testSelectControlSingle_label(self):
        import ClientForm
##         <SELECT name=year>
##          <OPTION value=0 label="2002">current year</OPTION>
##          <OPTION value=1>2001</OPTION>
##          <OPTION>2000</OPTION>
##         </SELECT>
        attrs = {"type": "ignored",
                 "name": "year",
                 "value": "0",
                 "label": "2002",
                 "contents": "current year"}
        attrs2 = {"type": "ignored",
                  "name": "year",
                  "value": "1",
                  "label": "2001",  # label defaults to contents
                  "contents": "2001"}
        attrs3 = {"type": "ignored",
                  "name": "year",
                  "value": "2000",  # value defaults to contents
                  "label": "2000",  # label defaults to contents
                  "contents": "2000"}
        c = ClientForm.SelectControl("select", attrs)
        c2 = ClientForm.SelectControl("select", attrs2)
        c3 = ClientForm.SelectControl("select", attrs3)
        c.merge_control(c2)
        c.merge_control(c3)
        c.fixup()

        assert c.possible_values() == ["0", "1", "2000"]
        assert c.possible_labels() == ["2002", "2001", "2000"]

        def bad_value(c=c): c.value = "2002"
        assert c.value == "0"
        c.value = None
        assert c.value == None
        c.value = "0"
        assert c.value == "0"
        c.value = None
        c.set_value_by_label("2002")
        assert c.value == "0"
        assert c.get_value_as_label() == "2002"
        c.set_value_by_label("2000")
        assert c.value == "2000"
        assert c.get_value_as_label() == "2000"

    def testSelectControlSingle(self):
        attrs = {"type": "this is ignored",
                 "name": "name_value",
                 "value": "value_value",
                 "label": "contents_value",
                 "contents": "contents_value"}
        # Netscape and IE behaviour...
        c = ClientForm.SelectControl("select", attrs)
        c.fixup()
        assert c.type == "select"
        assert c.name == "name_value"
        assert c.value == "value_value"
        assert c.possible_values() == ["value_value"]
        # ...and RFC 1866 behaviour are identical (unlike multiple SELECT).
        c = ClientForm.SelectControl("select", attrs, select_default=1)
        c.fixup()
        assert c.value == "value_value"

        # construct larger list from length-1 lists
        c = ClientForm.SelectControl("select", attrs)
        attrs2 = attrs.copy()
        attrs2["value"] = "value_value2"
        c2 = ClientForm.SelectControl("select", attrs2)
        c.merge_control(c2)
        c.fixup()
        assert str(c) == ("<SelectControl("
                           "name_value=[*value_value, value_value2])>")
        c.value = None
        assert c.value == None
        assert str(c) == ("<SelectControl("
                           "name_value=[value_value, value_value2])>")
        c.value = "value_value"
        assert c.value == "value_value"
        assert str(c) == ("<SelectControl("
                           "name_value=[*value_value, value_value2])>")
        assert c.possible_values() == ["value_value", "value_value2"]

        def bad_assign(c=c): c.value = ["value_value2"]
        self.assertRaises(AttributeError, bad_assign)
        c.value = "value_value2"
        assert c.value == "value_value2"
        c.value = "value_value"
        assert c.value == "value_value"
        def bad_assign(c=c): c.value = "oops"
        self.assertRaises(AttributeError, bad_assign)
        assert c.value == "value_value"
        self.assertRaises(TypeError, c.toggle, "value_value")
        assert c.value == "value_value"
        # nothing selected is allowed
        c.value = None
        assert c.value == None

    def testRadioControl(self):
        attrs = {"type": "this is ignored",
                 "name": "name_value",
                 "value": "value_value"}
        # Netscape and IE behaviour...
        c = ClientForm.RadioControl("radio", attrs)
        c.fixup()
        assert c.type == "radio"
        assert c.name == "name_value"
        assert c.value == None
        assert c.possible_values() == ["value_value"]
        # ...and RFC 1866 behaviour
        c = ClientForm.RadioControl("radio", attrs, select_default=1)
        c.fixup()
        assert c.value == "value_value"

        # construct larger list from length-1 lists
        c = ClientForm.RadioControl("radio", attrs, select_default=1)
        attrs2 = attrs.copy()
        attrs2["value"] = "value_value2"
        c2 = ClientForm.RadioControl("radio", attrs2, select_default=1)
        c.merge_control(c2)
        c.fixup()
        assert str(c) == ("<RadioControl("
                           "name_value=[*value_value, value_value2])>")
        assert c.possible_values() == ["value_value", "value_value2"]

        def list_assign(c=c): c.value = ["value_value2"]
        self.assertRaises(AttributeError, list_assign)
        c.value = "value_value2"
        assert c.value == "value_value2"
        c.value = "value_value"
        assert c.value == "value_value"
        def bad_assign(c=c): c.value = "oops"
        self.assertRaises(AttributeError, bad_assign)
        assert c.value == "value_value"
        self.assertRaises(TypeError, c.toggle, "value_value")
        assert c.value == "value_value"
        # nothing selected is allowed
        c.value = None
        self.assert_(c.value == None)


class FormTests(TestCase):
    base_uri = "http://auth.athensams.net/"
    def test_click(self):
        file = StringIO(
"""<form action="abc" name="myform">

<input type="submit" name="foo"></input>
</form>
""")
        form = ClientForm.ParseFile(file, "http://blah/")[0]
        self.assertRaises(ValueError, form.click, {"nr": 1})

    def testAuth(self):
        file = open("./testdata/Auth.html", "r")
        forms = ClientForm.ParseFile(file, self.base_uri)
        assert len(forms) == 1
        form = forms[0]
        assert form.action == (
            "http://auth.athensams.net/"
            "?ath_returl=%22http%3A%2F%2Ftame.mimas.ac.uk%2Fisicgi"
            "%2FWOS-login.cgi%22&ath_dspid=MIMAS.WOS")

        self.assertRaises(ValueError,
                          lambda form=form: form.toggle("oops", "d'oh"))
        self.assertRaises(IndexError, lambda form=form: form["oops"])
        def bad_assign(form=form): form["oops"] = "d'oh"
        self.assertRaises(IndexError, bad_assign)

        keys = ["ath_uname", "ath_passwd"]
        values = ["", ""]
        types = ["text", "password"]
        for i in range(len(keys)):
            key = keys[i]
            c = form.find_control(key)
            assert c.value == values[i]
            assert c.type == types[i]
        c = form.find_control(type="image")
        assert c.name == None
        assert c.value == None
        assert c.type == "image"

        form["ath_uname"] = "jbloggs"
        form["ath_passwd"] = "foobar"
        assert form.items() == [("ath_uname", "jbloggs"),
                                ("ath_passwd", "foobar")]

    def testSearchType(self):
        file = open("./testdata/SearchType.html", "r")
        forms = ClientForm.ParseFile(file, self.base_uri)
        assert len(forms) == 1
        form = forms[0]

        keys = ["SID", "SESSION_DIR", "Full Search", "Easy Search",
                "New Session", "Log off", "Form", "JavaScript"]
        values = ["PMrU0IJYy4MAAELSXic_E2011300_PMrU0IJYy4MAAELSXic-0",
                  "", None, None, None, None, "Welcome", "No"]
        types = ["hidden", "hidden", "image", "image", "image", "image",
                 "hidden", "hidden"]
        for i in range(len(keys)):
            key = keys[i]
            assert form.find_control(key).value == values[i]
            assert form.find_control(key).type == types[i]

        items = form.click("Full Search", items=1)
        assert items == [
            ("SID", "PMrU0IJYy4MAAELSXic_E2011300_PMrU0IJYy4MAAELSXic-0"),
            ("SESSION_DIR", ""),
            ("Full Search.x", "1"), ("Full Search.y", "1"),
            ("Form", "Welcome"), ("JavaScript", "No")]

    def testFullSearch(self):
        pass  # XXX

    def testGeneralSearch(self):
        file = open("./testdata/GeneralSearch.html", "r")
        forms = ClientForm.ParseFile(file, self.base_uri)
        assert len(forms) == 1
        form = forms[0]

        keys = ["SID", "SESSION_DIR",
                "Home", "Date & Database Limits", "Cited Ref Search",
                "Log off", "Search",
                "topic", "titleonly", "author", "journal", "address",
                "Search", "Save query", "Clear",
                "languagetype", "doctype", "Sort",
                "Form", "Func"]
        values = ["PMrU0IJYy4MAAELSXic_E2011300_PMrU0IJYy4MAAELSXic-0", "",
                  None, None, None, None, None,
                  "", [], "", "", "",
                  None, None, None,
                  ["All languages"], ["All document types"], "Latest date",
                  "General", "Search"]
        types = ["hidden", "hidden",
                 "image", "image", "image", "image", "image",
                 "text", "checkbox", "text", "text", "text",
                 "image", "image", "image",
                 "select", "select", "select",
                 "hidden", "hidden"]
        fc = form.find_control
        for i in range(len(keys)):
            name = keys[i]
            type = types[i]
            assert fc(name).value == values[i]
            assert fc(name).type == type
            assert fc(name, type).name == name
        assert fc(type="hidden").name == "SID"
        assert fc(type="image").name == "Home"
        assert fc(nr=6).name == "Search"
        assert fc(nr=50) == None
        self.assertRaises(ValueError, lambda fc=fc: fc(nr=-1))
        assert fc("Search", "image").name == "Search"
        assert fc("Search", "hidden") == None
        s0 = fc("Search", "image", 0)
        s0b = fc("Search", "image", 0)
        s1 = fc("Search", "image", 1)
        assert s0.name == s1.name == "Search"
        assert s0 is s0b
        assert s0 is not s1
        assert fc("Search", "image", 2) == None
        assert fc(type="text", nr=2).name == "journal"
        assert fc("Search", nr=0) is not fc("Search", nr=1)

        form["topic"] = "foo"
        assert form["topic"] == "foo"
        form["author"] = "bar"
        form["journal"] = ""
        form["address"] = "baz"
        form["languagetype"] = ["English", "Catalan"]
        assert form["languagetype"] == ["English", "Catalan"]
        form["titleonly"] = "on"
        assert form["titleonly"] == ["on"]
        items = form.click("Search", items=1)
        # XXX Add tests for items in ControlTests
        assert items == [
            ("SID", "PMrU0IJYy4MAAELSXic_E2011300_PMrU0IJYy4MAAELSXic-0"),
            ("SESSION_DIR", ""),
            ("Search.x", "1"), ("Search.y", "1"),
            ("topic", "foo"),
            ("titleonly", "on"),
            ("author", "bar"),
            ("journal", ""), ("address", "baz"),
            ("languagetype", "English"), ("languagetype", "Catalan"),
            ("doctype", "All document types"), ("Sort", "Latest date"),
            ("Form", "General"), ("Func", "Search")]

        pvs = form.possible_values("languagetype")
        assert pvs[0] == "All languages"
        assert len(pvs) == 47

        self.assertRaises(
            ValueError,
            lambda form=form: form.toggle("languagetype", "d'oh"))
        form.toggle("languagetype", "English")
        assert form["languagetype"] == ["Catalan"]

    def testResults(self):
        file = open("./testdata/Results.html", "r")
        forms = ClientForm.ParseFile(file, self.base_uri)
        assert len(forms) == 1
        form = forms[0]

        pvs = form.find_control("marked_list_candidates").possible_values()
        assert pvs == [
            "000174872000059/1", "000174858300003/2", "000174827900006/3"]
        def bad_setitem(form=form):
            form["marked_list_candidates"] = "blah"
        self.assertRaises(ValueError, bad_setitem)
        form["marked_list_candidates"] = pvs[0]

        # I've removed most of the INPUT elements from this page, and
        # corrected an HTML error
        keys = ["Add marked records to list",
                "Add records on page to list",
                "Add all records retrieved to list",
                "marked_list_candidates",
                "Add marked records to list",
                "Add records on page to list",
                "Add all records retrieved to list"
                ]
        types = ["image", "image", "image",
                 "checkbox",
                 "image", "image", "image"]
        values = [None, None, None,
                  [pvs[0]],
                  None, None, None
                 ]

        for i in range(len(keys)):
            key = keys[i]
            control = form.find_control(key)
            assert control.value == values[i]
            assert control.type == types[i]

        items = form.click("Add all records retrieved to list", items=1)
        assert items == [
            ("Add all records retrieved to list.x", "1"),
            ("Add all records retrieved to list.y", "1"),
            ("marked_list_candidates", pvs[0])]

    def testMarkedResults(self):
        file = open("./testdata/MarkedResults.html", "r")
        forms = ClientForm.ParseFile(file, self.base_uri)
        assert len(forms) == 1
        form = forms[0]

        items = form.click(items=1)
        # I've removed most of the INPUT elements from this page, and
        # corrected an HTML error
        assert items == [
            ("Add marked records to list.x", "1"),
            ("Add marked records to list.y", "1"),
            ("marked_list_candidates", "000174872000059/1"),
            ("marked_list_candidates", "000174858300003/2"),
            ("marked_list_candidates", "000174827900006/3")
            ]

    def testMarkedRecords(self):
        pass  # XXX


class MoreFormTests(TestCase):
    def make_form(self):
        f = StringIO("""\
<form blah="nonsense">
  <input type="checkbox" name="a" value="1" id="1" blah="spam"></input>
  <input type="checkbox" name="a" value="2" blah="eggs"></input>
  <input type="checkbox" name="a" value="3" id="3"></input>

  <select name="c" id="cselect" blah="foo">
    <option id="coption1" blah="bar">1</option>
    <option selected blah="baz">2</option>
    <option id="coption3">3</option>
  </select>
</form>
""")
        return ClientForm.ParseFile(f, "http://blah/")[0]

    def test_attrs(self):
        form = self.make_form()

        self.assert_(form.attrs["blah"] == "nonsense")

        a = form.find_control("a")
        self.assert_(not hasattr(a, "attrs"))
        self.assert_(a.find_item("1")["blah"] == "spam")
        self.assert_(a.find_item("2")["blah"] == "eggs")
        self.assert_(not a.find_item("3").has_key("blah"))

        c = form.find_control("c")
        # attrs dict is missing, fixed in 0.1.x
        self.assert_(not hasattr(c, "attrs"))
        self.assert_(c.find_item("1")["blah"] == "bar")
        self.assert_(c.find_item("2")["blah"] == "baz")
        # item 3 inherits attributes from parent SELECT element
        #  (this is fixed in 0.1.x)
        self.assert_(c.find_item("3")["blah"] == "foo")


if __name__ == "__main__":
    unittest.main()
